@@ -11,16 +11,19 @@ package Net::SIP::Leg;
 use Digest::MD5 'md5_hex';
 use Socket;
 use Net::SIP::Debug;
-use Net::SIP::Util qw( sip_hdrval2parts invoke_callback sip_uri_eq );
+use Net::SIP::Util ':all';
+use Net::SIP::SocketPool;
 use Net::SIP::Packet;
 use Net::SIP::Request;
 use Net::SIP::Response;
-use Errno 'EHOSTUNREACH';
+use Errno qw(EHOSTUNREACH EINVAL);
+use Hash::Util 'lock_ref_keys';
+use Carp;
 
-use fields qw( sock addr port proto contact branch via );
+use fields qw(contact branch via proto src socketpool);
 
 # sock: the socket for the leg
-# addr,port: addr,port where it listens
+# src: hash addr,port,family where it receives data and sends data from
 # proto: udp|tcp
 # contact: to identify myself (default from addr:port)
 # branch: base for branch-tag for via header
@@ -30,67 +33,181 @@ use fields qw( sock addr port proto contact branch via );
 # create a new leg
 # Args: ($class,%args)
 #   %args: hash, the following keys will be used and deleted from hash
-#      sock: socket, the addr,port and proto will be determined from this
-#      addr,port,proto: if sock is not given they will be used to
-#        create a socket. port defaults to 5060 and proto to udp
-#        if port is defined and 0 a port will be assigned from the system
-#      proto: defaults to udp
-#      contact: default based on addr and port
-#      branch: if not given will be created
-# Returns: $self
+#      proto: udp|tcp|tls. If not given will be determined from 'sock' or will
+#        default to 'udp' or 'tls' (if 'tls' arg is used)
+#      host,addr,port,family: source of outgoing and destination of
+#        incoming data.
+#        If IP address addr not given these values will be determined from
+#        'sock'. Otherwise port will default to 5060 or 5061 (tls) and family
+#        will be determined from addr syntax. host will default to addr
+#      dst: destination for this leg in case a fixed destination is used
+#        if not given 'sock' will be checked if connected
+#      sock: socket which can just be used
+#        if not given will create new socket based on proto, addr, port
+#        if dst is given this new socket will be connected (udp only)
+#      socketpool: socketpool which can just be used
+#        if not given a new SocketPool object will be created based on the given
+#        'sock' or the created socket (addr, port...). 'sock' and 'socketpool'
+#        must not be given both.
+#      tls: optional configuration parameters for IO::Socket::SSL. Implies
+#        use of proto 'tls'.
+#      contact: contact information
+#        default will be based on addr and port
+#      branch: branch informaton
+#        default will be based on proto, addr, port
+# Returns: $self - new leg object
 ###########################################################################
 sub new {
     my ($class,%args) = @_;
     my $self = fields::new($class);
 
-    if ( my $addr = delete $args{addr} ) {
+    my $proto = delete $args{proto};
+    my $dst = delete $args{dst};
+    my $tls = delete $args{tls};
+    $proto ||= 'tls' if $tls;
+
+    my ($sip_proto,$default_port) = $proto && $proto eq 'tls'
+	? ('sips',5061) : ('sip',5060);
+
+    my $family;
+    my $host = delete $args{host};
+    if (my $addr = delete $args{addr}) {
 	my $port = delete $args{port};
-	# port = 0 -> get port from system
-	if ( ! defined $port ) {
-	    $port = $1 if $addr =~s{:(\d+)$}{};
-	    $port ||= 5060;
+	my $family = delete $args{family};
+	if (!$family) {
+	    ($addr,my $port_a, $family) = ip_string2parts($addr);
+	    die "port given both as argument and contained in address"
+		if $port && $port_a && $port != $port_a;
+	    $port = $port_a if $port_a;
 	}
-	my $proto = $self->{proto} = delete $args{proto} || 'udp';
-	if ( ! ( $self->{sock} = delete $args{sock} ) ) {
-	    $self->{sock} = IO::Socket::INET->new(
-		Proto => $proto,
-		LocalPort => $port,
-		LocalAddr => $addr,
-	    ) || die "failed $proto $addr:$port $!";
-	}
-	if ( ! $port ) {
-	    # get the assigned port
-	    ($port) = unpack_sockaddr_in( getsockname( $self->{sock} ));
-	}
-
-	$self->{port} = $port;
-	$self->{addr} = $addr;
-
-    } elsif ( my $sock = $self->{sock} = delete $args{sock} ) {
-	# get data from socket
-	($self->{port}, my $addr) = unpack_sockaddr_in( $sock->sockname );
-	$self->{addr}  = inet_ntoa( $addr );
-	$self->{proto} = ( $sock->socktype == SOCK_STREAM ) ? 'tcp':'udp'
+	# port defined and 0 -> get port from system
+	$port = $default_port if ! defined $port;
+	$self->{src} = lock_ref_keys({
+	    host   => $host || $addr,
+	    addr   => $addr,
+	    port   => $port,
+	    family => $family
+	});
     }
 
-    my ($port,$sip_proto) =
-	$self->{port} == 5060 ? ( '','sip' ) :
-	( $self->{port} == 5061 and $self->{proto} eq 'tcp' ) ? ( '','sips' ) :
-	( ":$self->{port}",'sip' )
-	;
-    my $leg_addr = $self->{addr}.$port;
+    if ($dst && !ref($dst)) {
+	my ($ip,$port,$family) = ip_string2parts($dst);
+	$family or die "destination must contain IP address";
+	$dst = lock_ref_keys({
+	    host   => $ip,
+	    addr   => $ip,
+	    port   => $port,
+	    family => $family,
+	});
+    }
+
+    my $sock = delete $args{sock};
+    my $socketpool = delete $args{socketpool};
+    die "only socketpool or sock should be given" if $sock && $socketpool;
+    $sock ||= $socketpool && $socketpool->master;
+
+    my $sockpeer = undef;
+    if (!$sock) {
+	# create new socket
+	$proto ||= 'udp';
+	my $src = $self->{src};
+	if (!$src) {
+	    # no src given, try to get useable soure from dst
+	    die "neither source, destination nor socket given" if !$dst;
+	    my $srcip = laddr4dst($dst->{addr}) or die
+		"cannot find local IP when connecting to $dst->{addr}";
+	    $src = $self->{src} = lock_ref_keys({
+		host   => $host || $srcip,
+		addr   => $srcip,
+		port   => 0,
+		family => $dst->{family},
+	    });
+	}
+
+	croak("addr must be IP address") if ! ip_is_v46($src->{addr});
+
+	my %sockargs = (
+	    Proto     => $proto eq 'tls' ? 'tcp' : $proto,
+	    Family    => $src->{family},
+	    LocalAddr => $src->{addr},
+	    Reuse     => 1, ReuseAddr => 1,
+	);
+	if ($proto eq 'tcp' or $proto eq 'tls') {
+	    # with TCP we create a listening socket
+	    $sockargs{Listen} = 100;
+	} elsif ($dst) {
+	    # with UDP we can create a connected socket if dst is given
+	    $sockargs{PeerAddr} = $dst->{addr};
+	    $sockargs{PeerPort} = $dst->{port};
+	    $sockpeer = $dst;
+	}
+
+	# create a socket with the given local port
+	# if no port is given try 5060,5062.. or let the system pick one
+	for my $port ($src->{port}
+	    ? $src->{port}
+	    : ($default_port, 5062..5100, 0)) {
+	    last if $sock = INETSOCK(%sockargs, LocalPort => $port);
+	}
+
+	$sock or die "failed to bind to " . ip_parts2string($src).": $!";
+	$src->{port} ||= $sock->sockport;
+	DEBUG(90,"created socket on ".ip_parts2string($src));
+
+    } else {
+	# get proto from socket
+	$proto ||= $sock->socktype == SOCK_DGRAM ? 'udp':'tcp';
+
+	# get src from socket
+	if (!$self->{src}) {
+	    my $saddr = getsockname($sock) or die
+		"cannot get local name from provided socket: $!";
+	    $self->{src} = ip_sockaddr2parts($saddr);
+	    $self->{src}{host} = $host if $host;
+	}
+	if (!$dst and my $saddr = getpeername($sock)) {
+	    # set dst from connected socket
+	    $sockpeer = $dst = ip_sockaddr2parts($saddr);
+	}
+    }
+
+    # create socketpool and add primary socket of leg to it if needed
+    $self->{socketpool} = $socketpool ||= Net::SIP::SocketPool->new(
+	$proto, $sock, $dst, $sockpeer, $tls);
+
+    my $leg_addr = ip_parts2string({
+	%{$self->{src}},
+	use_host => 1, # prefer hostname
+	default_port => $default_port,
+    }, 1);  # use "[ipv6]" even if no port is given
     $self->{contact}  = delete $args{contact} || "$sip_proto:$leg_addr";
 
-    $self->{branch} = 'z9hG4bK'.
-	( delete $args{branch} || md5_hex( @{$self}{qw( addr port proto )} ));
+    $self->{branch} = 'z9hG4bK'. (
+	delete $args{branch}
+	|| md5_hex(@{$self->{src}}{qw(addr port)}, $proto)  # ip, port, proto
+    );
 
-    $self->{contact} =~m{^\w+:(.*)};
     $self->{via} =  sprintf( "SIP/2.0/%s %s;branch=",
-	uc($self->{proto}),$leg_addr );
+	uc($proto),$leg_addr );
+    $self->{proto} = $proto;
+
+    die "unhandled arguments: ".join(", ", keys %args) if %args;
 
     return $self;
 }
 
+###########################################################################
+# do we need retransmits on this leg?
+# Args: $self
+# Returns: 1|0
+#   1: need retransmits (UDP)
+#   0: don't need retransmits (TCP, TLS)
+###########################################################################
+sub do_retransmits {
+    my Net::SIP::Leg $self = shift;
+    return $self->{proto} eq 'udp' ? 1 : 0;
+}
+
 ###########################################################################
 # prepare incoming packet for forwarding
 # Args: ($self,$packet)
@@ -200,6 +317,9 @@ sub forward_outgoing {
 	    my $branch = $self->via_branch($packet,3);
 	    foreach my $via ( @via ) {
 		my (undef,$param) = sip_hdrval2parts( via => $via );
+		# ignore via header w/o branch, although these don't conform to
+		# RFC 3261, sect 8.1.1.7
+		defined $param->{branch} or next;
 		if ( substr( $param->{branch},0,length($branch) ) eq $branch ) {
 		    DEBUG( 10,'loop detected because outgoing leg is in Via. DROP' );
 		    return [ undef,'loop detected on outgoing leg, dropping' ];
@@ -244,17 +364,17 @@ sub forward_outgoing {
 ###########################################################################
 # deliver packet through this leg to specified addr
 # add local Via header to requests
-# Args: ($self,$packet,$addr;$callback)
+# Args: ($self,$packet,$dst;$callback)
 #   $packet: Net::SIP::Packet
-#   $addr:   ip:port where to deliver
-#   $callback: optional callback, if an error occured the callback will
-#      be called with $! as argument. If no error occured and the
+#   $dst:    target for delivery as hash host,addr,port,family
+#   $callback: optional callback, if an error occurred the callback will
+#      be called with $! as argument. If no error occurred and the
 #      proto is tcp the callback will be called with error=0 to show
-#      that the packet was definitly delivered (and need not retried)
+#      that the packet was definitely delivered (and there's no need to retry)
 ###########################################################################
 sub deliver {
     my Net::SIP::Leg $self = shift;
-    my ($packet,$addr,$callback) = @_;
+    my ($packet,$dst,$callback) = @_;
 
     my $isrq = $packet->is_request;
     if ( $isrq ) {
@@ -301,112 +421,61 @@ sub deliver {
 	$packet->insert_header( supported => '' );
     }
 
+    die "target protocol $dst->{proto} does not match leg $self->{proto}"
+	if exists $dst->{proto} && $dst->{proto} ne $self->{proto};
+    $dst->{port} ||= $self->{proto} eq 'tls' ? 5061 : 5060;
 
-    my ($proto,$host,$port) =
-	$addr =~m{^(?:(\w+):)?([\w\-\.]+)(?::(\d+))?$};
-    #DEBUG( "%s -> %s %s %s",$addr,$proto||'',$host, $port||'' );
-    $port ||= $proto eq 'sips' ? 5061: 5060;
-
-
-    $self->sendto( $packet->as_string, $host,$port,$callback )
-	|| return;
-    DEBUG( 2, "delivery from $self->{addr}:$self->{port} to $addr OK:\n%s",
+    $DEBUG && DEBUG( 2, "delivery with %s from %s to %s:\n%s",
+	$self->{proto},
+	ip_parts2string($self->{src}),
+	ip_parts2string($dst),
 	$packet->dump( Net::SIP::Debug->level -2 ) );
+
+    return $self->sendto($packet,$dst,$callback);
 }
 
 ###########################################################################
 # send data to peer
-# Args: ($self,$data,$host,$port,$callback)
-#   $data: string representation of SIP packet
-#   $host: target ip
-#   $port: target port
+# Args: ($self,$packet,$dst,$callback)
+#   $packet: SIP packet object
+#   $dst:   target as hash host,addr,port,family
 #   $callback: callback for error|success, see method deliver
 # Returns: $success
-#   $success: true if no problems occured while sending (this does not
+#   $success: true if no problems occurred while sending (this does not
 #     mean that the packet was delivered reliable!)
 ###########################################################################
 sub sendto {
     my Net::SIP::Leg $self = shift;
-    my ($data,$host,$port,$callback) = @_;
+    my ($packet,$dst,$callback) = @_;
 
-    # XXXXX for now udp only
-    # for tcp the delivery might be done over multiple callbacks
-    # (eg whenever I can write on the socket)
-    # for tcp I need to handle the case where I got a request on
-    # the leg, then the leg got closed and the I've need to deliver
-    # the response over a new leg, created based on the master leg
-    # eg I still need to know which outgoing master leg I have,
-    # even if my real outgoing leg is closed (responsed might be
-    # delivered over the same tcp connection, but no need to do so)
-
-    if ( $self->{proto} ne 'udp' ) {
-	use Errno 'EINVAL';
-	DEBUG( 1,"can only proto udp for now, but not $self->{proto}" );
-	invoke_callback( $callback, EINVAL );
-    }
-
-    my $host4 = inet_aton( $host ) or do {
-	# this should not happen because host should better be IP
-	DEBUG( 1, "lookup problems of $host?" );
-	invoke_callback( $callback, EINVAL );
-	return;
-    };
-
-    my $target = sockaddr_in( $port,$host4 );
-    unless ( $self->{sock}->send( $data,0,$target )) {
-	DEBUG( 1,"send failed: callback=$callback error=$!" );
-	invoke_callback( $callback, $! );
-	return;
-    }
-
-    # XXXX dont forget to call callback back with error=0 if
-    # delivery by tcp successful
-    return 1;
+    $self->{socketpool}->sendto($packet,$dst,$callback)
+	&& return 1;
+    return;
 }
 
 ###########################################################################
-# receive packet
-# for udp socket it just makes a recv on the socket and returns the packet
-# for tcp master sockets it makes accept and creates a new leg based on
-#   the masters leg.
-# Args: ($self)
-# Returns: ($packet,$from) || ()
-#   $packet: Net::SIP::Packet
-#   $from:   ip:port where it got packet from
+# Handle newly received packet.
+# Currently just passes through the packet
+# Args: ($self,$packet,$from)
+#   $packet: packet object
+#   $from: hash with proto,addr,port,family where the packet came from
+# Returns: ($packet,$from)|()
+#   $packet: packet object
+#   $from: hash with proto,ip,port,family where the packet came from
 ###########################################################################
 sub receive {
     my Net::SIP::Leg $self = shift;
+    my ($packet,$from) = @_;
 
-    if ( $self->{proto} ne 'udp' ) {
-	DEBUG( 1,"only udp is supported at the moment" );
-	return;
-    }
-
-    my $from = recv( $self->{sock}, my $buf, 2**16, 0 ) or do {
-	DEBUG( 1,"recv failed: $!" );
-	return;
-    };
-
-    # packet must be at least 13 bytes big (first line incl version
-    # + final crlf crlf). Ignore anything smaller, probably keep-alives
-    if ( length($buf)<13 ) {
-	DEBUG(11,"ignored packet with len ".length($buf)." because to small (keep-alive?)");
-	return;
-    }
-
-    my $packet = eval { Net::SIP::Packet->new( $buf ) } or do {
-	DEBUG( 3,"cannot parse buf as SIP: $@\n$buf" );
-	return;
-    };
-
-    my ($port,$host) = unpack_sockaddr_in( $from );
-    $host = inet_ntoa($host);
-    DEBUG( 2,"received on $self->{addr}:$self->{port} from $host:$port packet\n%s",
-	$packet->dump( Net::SIP::Debug->level -2 ));
-
-    return ($packet,"$host:$port");
+    $DEBUG && DEBUG( 2,"received packet on %s from %s:\n%s",
+	sip_sockinfo2uri($self->{proto},@{$self->{src}}{qw(addr port family)}),
+	sip_sockinfo2uri(@{$from}{qw(proto addr port family)}),
+	$packet->dump( Net::SIP::Debug->level -2 )
+    );
+    return ($packet,$from);
 }
 
+
 ###########################################################################
 # check if the top via header matches the transport of this call through
 # this leg. Used to strip Via header in response.
@@ -448,15 +517,30 @@ sub via_branch {
     my ($packet,$level) = @_;
     my $val = $self->{branch};
     $val .= substr( md5_hex( $packet->tid ),0,15 ) if $level>1;
-    $val .= substr( md5_hex( 
-	( sort $packet->get_header( 'proxy-authorization' )),
-	( sort $packet->get_header( 'proxy-require' )),
-	$packet->get_header( 'route' ),
-	$packet->get_header( 'to' ),
-	$packet->get_header( 'from' ),
-	($packet->get_header( 'via' ))[0] || '',
-	($packet->as_parts())[1],
-    ),0,15 ) if $level>2;
+    if ($level>2) {
+	my @parts;
+	# RT#120816 -  take only known constant values from proxy-authorization
+	for(sort $packet->get_header('proxy-authorization')) {
+	    my ($typ,$param) = sip_hdrval2parts('proxy-authorization' => $_);
+	    push @parts,$typ;
+	    for(qw(realm username domain qop algorithm)) {
+		push @parts,"$_=$param->{$_}" if exists $param->{$_};
+	    }
+	}
+
+	# RT#120816 - include only the branch from via header if possible
+	if (my $via = ($packet->get_header('via'))[0]) {
+	    my (undef,$param) = sip_hdrval2parts(via => $via);
+	    push @parts, $param && $param->{branch} || $via;
+	}
+
+	push @parts,
+	    ( sort $packet->get_header('proxy-require')),
+	    $packet->get_header('route'),
+	    $packet->get_header('from'),
+	    ($packet->as_parts())[1]; # URI
+	$val .= substr(md5_hex(@parts),0,15);
+    }
     return $val;
 }
 
@@ -472,18 +556,16 @@ sub can_deliver_to {
     my Net::SIP::Leg $self = shift;
     my %spec;
     if (@_>1) {
-	%spec = @_
+	%spec = @_;
     } else {
-	my $spec = shift;
-	my ($proto,$addr) = $spec =~m{^(?:(udp|tcp):)?([^:]+)}
-	    or return; # wrong spec?
-	$spec{proto} = $proto if $proto;
-	$spec{addr}  = $addr;
-	# ignore port
+	@spec{ qw(proto host port family) } = sip_uri2sockinfo(shift());
+	$spec{addr} = $spec{family} ? $spec{host} : undef;
     }
 
-    # check against proto of leg
-    return if ( $spec{proto} && $spec{proto} ne $self->{proto} );
+    # return false if proto or family don't match
+    return if $spec{proto} && $spec{proto} ne $self->{proto};
+    return if $spec{family} && $self->{src}
+	&& $self->{src}{family} != $spec{family};
 
     # XXXXX dont know how to find out if I can deliver to this addr from this
     # leg without lookup up route
@@ -493,13 +575,53 @@ sub can_deliver_to {
 }
 
 ###########################################################################
-# returns FD on Leg
-# Args: $self
-# Returns: socket of leg
+# check if this leg matches given criteria (used in Dispatcher)
+# Args: ($self,$args)
+#   $args: hash with any of 'addr', 'port', 'proto', 'sub'
+# Returns: true if leg fits all args
 ###########################################################################
-sub fd {
+sub match {
     my Net::SIP::Leg $self = shift;
-    return $self->{sock};
+    my $args = shift;
+    return if $args->{addr}
+	&& $args->{addr} ne $self->{src}{addr}
+	&& $args->{addr} ne $self->{src}{host};
+    return if $args->{port}  && $args->{port}  != $self->{src}{port};
+    return if $args->{proto} && $args->{proto} ne $self->{proto};
+    return if $args->{sub}   && !invoke_callback($args->{sub},$self);
+    return 1;
+}
+
+###########################################################################
+# returns SocketPool object on Leg
+# Args: $self
+# Returns: $socketpool
+###########################################################################
+sub socketpool {
+    my Net::SIP::Leg $self = shift;
+    return $self->{socketpool};
+}
+
+###########################################################################
+# local address of the leg
+# Args: $self;$parts
+#  $parts: number of parts to include
+#     0 -> address only
+#     1 -> address[:non_default_port]
+#     2 -> host[:non_default_port]
+# Returns: string
+###########################################################################
+sub laddr {
+    my Net::SIP::Leg $self = shift;
+    my $parts = shift;
+    ! $parts and return $self->{src}{addr};
+    return ip_parts2string({
+	%{ $self->{src} },
+	default_port => $self->{proto} eq 'tls' ? 5061 : 5060,
+	$parts == 1 ? () :
+	$parts == 2 ? (use_host => 1) :
+	die "invalid parts specification $parts",
+    });
 }
 
 ###########################################################################
@@ -509,7 +631,8 @@ sub fd {
 ###########################################################################
 sub dump {
     my Net::SIP::Leg $self = shift;
-    return ref($self)." $self->{proto}:$self->{addr}:$self->{port}";
+    return ref($self)." $self->{proto}:"
+	. ip_parts2string($self->{src});
 }
 
 
@@ -520,7 +643,8 @@ sub dump {
 ###########################################################################
 sub key {
     my Net::SIP::Leg $self = shift;
-    return "$self->{proto}:$self->{addr}:$self->{port}";
+    return ref($self).' '.join(':',$self->{proto},
+	@{$self->{src}}{qw(addr port)});
 }
 
 1;
